"""
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
"""


import collections
import json

import boto3
from disdat import api
import logging
from typing import Any, Union
import json


LOG_LEVEL = logging.INFO + 1


class PathParam:
    FULL_PARAM = '_full_params'
    CACHE_PARAM = '_cache_params'
    DSDT_PASS_PARAM = '_dsdt_params'
    DSDT_ONLY_ARGS = '_dsdt_only_args'
    USE_CACHE = '_use_cache'
    CACHE_DATA = '_data'

    DSDT_PASS_PARAM_SUFFIX = '{}.$'.format(DSDT_PASS_PARAM)
    DSDT_ONLY_ARGS_SUFFIX = '{}.$'.format(DSDT_ONLY_ARGS)

    FULL_PARAM_PREFIX = '$.{}'.format(FULL_PARAM)
    FULL_PARAM_SUFFIX = '{}.$'.format(FULL_PARAM)

    CACHE_PARAM_PREFIX = '$.{}'.format(CACHE_PARAM)
    CACHE_PARAM_SUFFIX = '{}.$'.format(CACHE_PARAM)

    USE_CACHE_PREFIX = '$.{}'.format(USE_CACHE)


class Cache:

    def __init__(self, dsdt_args):
        """
        Class that implements cache-pull and cache-push
        :param dsdt_args: dict, parameters necessary for disdat to set up and talk to remote contexts
            the params are supplied via the definition json (users don't need to do this as Caching.cache_step handles
            this)
        """
        self.context = dsdt_args['context']
        self.s3_url = dsdt_args['s3_bucket_url']
        self.bundle_name = dsdt_args['bundle_name']
        self.force_rerun = dsdt_args.get('force_rerun', False)
        self.verbose = dsdt_args.get('verbose', False)
        self.delocalize = dsdt_args.get('delocalized', False)
        self.state_machine_name = dsdt_args.get('state_machine_name', '')

        if self.verbose:
            level = LOG_LEVEL
        else:
            level = LOG_LEVEL + 1
        if len(logging.getLogger().handlers) > 0:
            logging.getLogger().setLevel(level=level)
        else:
            logging.basicConfig(format='%(asctime)s %(message)s', level=level)
        # set up local context
        api.context(self.context)
        # set up and bind with the remote context
        api.remote(self.context, remote_context=self.context, remote_url=self.s3_url)

    def cache_pull(self, event: Any) -> dict:
        """
        pulls data from the remote context specified by the constructor. I
        :param event: event data from the lambda function
        :return: dict, {PathParam.FULL_PARAM: Any, full data from the successor ,
                        PathParam.CACHE_PARAM: Any, what parameters to cache ,
                        PathParam.CACHE_DATA: Any, not null if a cache hit,
                        PathParam.USE_CACHE: bool, flag for the choice state }
        """
        full_params = event[PathParam.FULL_PARAM]
        cache_params = event[PathParam.CACHE_PARAM]
        use_cache, cached_data = False, None
        # since all states share the same lambda, we need a better identifier for the logs
        func_name = 'cache_pull_4_{}'.format(self.bundle_name)
        logging.log(level=LOG_LEVEL, msg='{} - received input event {}'.format(func_name, event))

        if not self.force_rerun:
            signature = {'input_params': json.dumps(cache_params)}
            # uniquely determine a proc name based on bundle name and signature
            proc_name = api.Bundle.calc_default_processing_name(self.bundle_name, signature, dep_proc_ids={})
            # pull bundle meta data from s3
            api.pull(self.context, self.bundle_name)
            # search if the proc name exists
            bundle = api.search(self.context, processing_name=proc_name)
            # could have multiple hits because of forced reruns
            if len(bundle) > 0:
                # use the latest cache data
                latest_bundle = bundle[0]
                # check if the signature match
                logging.log(level=LOG_LEVEL, msg='{} - {} bundles found with name {}'.format(func_name,
                                                                                              len(bundle),
                                                                                              self.bundle_name))
                use_cache = True not in [v != latest_bundle.params.get(k, None)
                                         for k, v in signature.items()]
                # if use cache is true, pulls the actual data (the json file that holds the cached data) from s3
                if use_cache:
                    api.pull(self.context, uuid=latest_bundle.uuid, localize=True)
                    file = latest_bundle.data
                    with open(file, 'r') as fp:
                        cached_data = json.load(fp)
        # return the result, full param is what the user step expects to receive, so we need to forward it
        # cache_params is needed by cache push to create bundles
        # cached_data is needed by cache push if use_cache is true
        # use_cache is needed by the choice state
        data = {PathParam.FULL_PARAM: full_params, PathParam.CACHE_PARAM: cache_params,
                PathParam.CACHE_DATA: cached_data, PathParam.USE_CACHE: use_cache}
        logging.log(level=LOG_LEVEL, msg='{} - outputs - {}'.format(func_name, data))
        return data

    def cache_push(self, event: Any, parent: Union[None, api.Bundle] = None) -> Any:
        """
        pushes data to the remote context is use cache is false
        otherwise simply parse the input event and return the cached data
        :param event: Any, whatever data received by the lambda function
        :return: Any, data output by the user step
        """

        logging.log(level=LOG_LEVEL, msg='found parent {}'.format(parent.uuid if parent else None))

        full_params = event[PathParam.FULL_PARAM]
        func_name = 'cache_push_4_{}'.format(self.bundle_name)
        logging.log(level=LOG_LEVEL, msg='{} - received input event {}'.format(func_name, event))

        # if the input event has type dict, it must comes from caching pull directly (use cache = true)
        if isinstance(full_params, dict):
            logging.log(level=LOG_LEVEL, msg='{} - no push'.format(func_name))
            # simply grab the cached data and return
            return full_params[PathParam.CACHE_DATA]

        # if the input is a list, then we know it's from the execution branch
        elif isinstance(full_params, list):
            # get the parameters that the user step used to generate output
            if PathParam.CACHE_PARAM in full_params[0]:
                cache_params = full_params[0][PathParam.CACHE_PARAM]
                params_to_save = full_params[1]
            elif PathParam.CACHE_PARAM in full_params[1]:
                cache_params = full_params[1][PathParam.CACHE_PARAM]
                params_to_save = full_params[0]
            else:
                raise ValueError('Key _cache_params is expected but not present')
            # create bundle signature
            signature = {'input_params': json.dumps(cache_params)}
            proc_name = api.Bundle.calc_default_processing_name(self.bundle_name, signature, dep_proc_ids={})
            with api.Bundle(self.context, name=self.bundle_name, processing_name=proc_name) as b:
                # write the output data to a json file. This will speed up caching pull
                # because we only pull the actual data after a signature match
                file = b.get_file('cached_data.json')
                with open(file, 'w') as f:
                    json.dump(params_to_save, f)
                b.add_params(signature)
                b.add_data(file)
                if parent is not None:
                    b.add_dependencies(parent)

            # commit and push the data to the remote context
            api.commit(self.context, self.bundle_name)
            api.push(self.context, bundle_name=self.bundle_name, delocalize=self.delocalize)
            logging.log(level=LOG_LEVEL,
                        msg='{} - data pushed. Cached parameters: {}, cached data: {}'\
                        .format(func_name, cache_params, params_to_save))
            return params_to_save

        else:
            raise TypeError("field {} must have type dict or list; {} is provided".format(
                PathParam.FULL_PARAM, type(full_params)))

    def get_lineage(self):
        logging.log(level=LOG_LEVEL, msg='run lineage tracing for {}'.format(self.state_machine_name))
        logs = self._get_execution_history_by_name(self.state_machine_name)
        parent_exec = self.lineage_tracer(logs, self.bundle_name)
        return self._get_bundle(parent_exec)

    def _get_execution_history_by_name(self, workflow_name):
        client = boto3.client('stepfunctions')
        workflow = []
        try:
            response = client.list_state_machines(maxResults=100)
            workflow += response['stateMachines']
            while 'nextToken' in response:
                response = client.list_state_machines(maxResults=100, nextToken=response['nextToken'])
                workflow += response['stateMachines']

            state_machine = [w for w in workflow if w['name'] == workflow_name]
            sm_arn = state_machine[0]['stateMachineArn']

            curr_execs = client.list_executions(stateMachineArn=sm_arn, statusFilter='RUNNING')['executions']
            assert len(curr_execs) <= 1, 'lineage capture may be incorrect if you have ' \
                                         'current executions, disabling lineage'
            logs = client.get_execution_history(executionArn=curr_execs[0]['executionArn'],
                                                maxResults=1000,
                                                reverseOrder=True)
            start = True
            while 'nextToken' in logs or start:
                if start:
                    logs = client.get_execution_history(executionArn=curr_execs[0]['executionArn'],
                                                        maxResults=1000,
                                                        reverseOrder=True)
                else:
                    logs = client.get_execution_history(executionArn=curr_execs[0]['executionArn'],
                                                        maxResults=1000,
                                                        reverseOrder=True,
                                                        nextToken=logs['nextToken'])
                start = False
                yield logs['events']

        except Exception as e:
            logging.warning('lambda failed to retrieve execution log. Check your lambda permission!')
            logging.warning('Error log ' + str(e))
            yield None

    @classmethod
    def lineage_tracer(cls, logs, bundle_name: str) -> Union[None, 'ExecutionEvent']:
        target_name = 'cache_pull_{}'.format(bundle_name)
        target_node = None
        events_by_id = {}
        children_by_id = collections.defaultdict(set)

        for page in logs:
            if page is None:
                return None
            # build partial DAG
            for event in page:
                event_id, parent_id = event['id'], event['previousEventId']
                children_by_id[parent_id].add(event_id)

                if event_id not in event:
                    events_by_id[event_id] = ExecutionEvent(event)
                event_obj = events_by_id[event_id]
                if event_id in children_by_id:
                    for child_id in children_by_id[event_id]:
                        if events_by_id[child_id].parent is None:
                            events_by_id[child_id].parent = event_obj
                            print(events_by_id[child_id])
            # find the target node
            for node in events_by_id.values():
                if node.name is not None and node.name == target_name:
                    target_node = node
                    break
            # target node not found in this page, go to the next page
            if target_node is None:
                continue

            parent_id = target_node.find_prev_bundle()
            if parent_id is None:
                continue
            else:
                return events_by_id[parent_id]
        # parent bundle not found
        # print([str(event_obj) for event_obj in events_by_id.values()])
        return None

    def _get_bundle(self, execution) -> Union[None, api.Bundle]:
        if execution is None:
            return None
        parent_bundle_name = execution.bundle_name
        output_params = json.loads(execution.full_event['stateExitedEventDetails']['output'])
        cache_params = output_params[PathParam.CACHE_PARAM]
        signature = {'input_params': json.dumps(cache_params)}
        proc_name = api.Bundle.calc_default_processing_name(parent_bundle_name, signature, dep_proc_ids={})
        api.pull(self.context, parent_bundle_name)
        bundles = api.search(self.context, parent_bundle_name, processing_name=proc_name)
        return bundles[0] if len(bundles) > 0 else None


class ExecutionEvent:
    def __init__(self, event: dict):
        self.full_event = event
        self.id = self.full_event['id']
        self.name = None
        self.parent = None
        self.bundle_name = None
        # self.children = None

        if 'stateExitedEventDetails' in self.full_event and 'name' in self.full_event['stateExitedEventDetails']:
            name = self.full_event['stateExitedEventDetails']['name']
            if name.startswith('cache_pull'):
                self.name = name
                self.bundle_name = self.name[len('cache_pull_'):]

    def find_prev_bundle(self) -> Union[None, str]:
        parent = self.parent
        counter = 0
        while parent is not None:

            if parent.bundle_name is not None:
                counter += 1
                if counter > 1:
                    return parent.id
            else:
                parent = parent.parent
        return None

    # def find_next_bundle(self) -> Union[None, set]:
    #
    #     children = self.children

    def __str__(self):
        return '{} -> {}; Name={}; bundle={}'.format(self.id, self.parent.id if self.parent else None,
                                                     self.name, self.bundle_name)


"""
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
"""


import pathlib
import time
from typing import Union
from stepfunctions import steps, inputs
from stepfunctions.steps import states
from stepfunctions.steps.fields import Field
import os
import shutil
from disdat_step_function.cache_lambda import PathParam as pp
import logging


class Caching:

    def __init__(self,
                 caching_lambda_name: str,
                 s3_bucket_url: str,
                 context_name: str,
                 state_machine_name: str = '',
                 force_rerun: bool = False,
                 verbose: bool = False):
        """
        This class initializes a caching object that contains basic specs of the caching layer
        :param caching_lambda_name: str, name of the lambda function. For instance 'caching-lambda'
        :param s3_bucket_url: str, link to the s3 bucket, in the format of 's3://BUCKET_NAME'
        :param context_name: str, the name of the context in which data versions are managed
        :param force_rerun: bool, rerun states by force
        :param verbose: bool, true to see detailed logs from the caching code
        """
        self.caching_lambda = caching_lambda_name
        self.s3_bucket = s3_bucket_url
        self.context_name = context_name
        self.force_rerun = force_rerun
        self.state_machine_name = state_machine_name
        self.verbose = verbose
        # kwargs passed to the caching lambda, users don't need to worry about this
        self.disdat_args = {'s3_bucket_url': self.s3_bucket,
                            'context': self.context_name,
                            'force_rerun': self.force_rerun,
                            'verbose': self.verbose,
                            'state_machine_name': self.state_machine_name}
        assert self.s3_bucket.startswith('s3://'), 's3 bucket url invalid format'
        assert isinstance(self.verbose, bool), 'verbose has the wrong type, bool expected'
        assert isinstance(self.force_rerun, bool), 'force_rerun has the wrong type, bool expected'

    def cache_step(self, user_step: steps.states, bundle_name: str = None, force_rerun: bool = None) -> steps.Chain:
        """
        enable caching for the input user step by wrapping the user step in a chain of generated caching states
        for instance:
            input: task A
            output: chain that contains caching pull -> choice ->(rerun)-> run task A -> caching push
                                                               |-> (no rerun) --------|

        :param user_step: steps.states, a stepfunction state object
        :param bundle_name: str, the unified bundle name for all data generated by this user step
        :param force_rerun: bool, override the object-level force rerun setting
        :return: steps.Chain, a mini DAG that implements the caching logic
        """
        task_name = user_step.state_id.replace(' ', '_').lower()
        # bundle name is automatically assigned if not set
        if bundle_name is None:
            bundle_name = task_name
        self.disdat_args['bundle_name'] = bundle_name
        # copy the dict parameters because it is shared by potentially many cache_step calls.
        disdat_args = self.disdat_args.copy()
        # override the force_rerun logic if specified
        if force_rerun is not None:
            disdat_args['force_rerun'] = force_rerun
        disdat_args['time'] = time.time()
        # set the caching pull input path to match user step's input path, we do this to avoid
        # caching unnecessary params that are not consumed by user step
        user_inputs = user_step.fields.get(Field.InputPath.value, '$')
        # define caching pull state
        cache_pull = steps.LambdaStep(state_id='cache_pull_{}'.format(task_name),
                                      output_path='$.Payload',
                                      parameters={
                                          'FunctionName': self.caching_lambda,
                                          'Payload': {pp.FULL_PARAM_SUFFIX: '$',
                                                      pp.CACHE_PARAM_SUFFIX: user_inputs,
                                                      pp.DSDT_ONLY_ARGS: disdat_args
                                                      }
                                        }
                                      )
        # define caching push state
        # note that the two states share the same lambda function!
        # as a matter of fact, this one lambda is called repeated by all cached states
        cache_push = steps.LambdaStep(state_id='cache_push_{}'.format(task_name),
                                      output_path='$.Payload',
                                      parameters={
                                          'FunctionName': self.caching_lambda,
                                          'Payload': {pp.FULL_PARAM_SUFFIX: '$',
                                                      pp.DSDT_ONLY_ARGS: disdat_args
                                                      }
                                        }
                                      )

        cache_condition = steps.Choice(state_id='use_cache?_{}'.format(task_name))
        # define the execution branch (run when input param signature does not match cache)
        execution_branch = steps.Parallel(state_id='execute_{}'.format(task_name))
        # we need a pass state to pass the inputs of the user step to caching push
        # otherwise we cannot recover the inputs based on user step's output
        execution_branch.add_branch(steps.Pass('param_pass_{}'.format(task_name),
                                               input_path=pp.CACHE_PARAM_PREFIX,
                                               parameters={pp.CACHE_PARAM_SUFFIX: '$'}))
        # because the output of caching pull has the following format
        # {'full param': ..., 'cache param': ..., 'use_cache':...}
        # it is important that we pass only the full_param field to the user step as expected by the user
        # The caching layer should be invisible
        param_resolve = steps.Pass('param_resolve_{}'.format(task_name), input_path=pp.FULL_PARAM_PREFIX)
        # param resolver is ran before the user step to properly prepare the input data
        execution_branch.add_branch(steps.Chain([param_resolve, user_step]))
        execution_branch = steps.Chain([execution_branch, cache_push])
        # if 'use_cache' is false, run caching push (it does not actually push, just simply parse and return the data)
        # if 'use_cache' is true, run the execution branch as well as caching push (create new version in ctxt)
        cache_condition.add_choice(rule=steps.ChoiceRule.BooleanEquals(pp.USE_CACHE_PREFIX, value=False),
                                   next_step=execution_branch)

        return steps.Chain([cache_pull, cache_condition, cache_push])


class PipelineCaching:

    def __init__(self, definition: Union[states.Chain, states.State], caching: Caching):
        """
        Cache all tasks in a pre-existing state machine
        :param definition: states.Chain, the state machine that the user wants to cache
        :param caching: Caching, the caching utility object used for caching individual states
        """
        visitor = StateVisitor(caching)
        definition.accept(visitor)

        self.definition = definition
        self.states = visitor.states
        # to replace keeps track of tasks to replace (determined by StateVisitor on the fly)
        self.to_replace = visitor.to_replace
        self.visited = {}

        if len(logging.getLogger().handlers) > 0:
            logging.getLogger().setLevel(level=logging.WARNING)
        else:
            logging.basicConfig(format='%(asctime)s %(levelname)s: %(message)s', level=logging.WARNING)

    def cache(self):
        """
        Cache the state machine passed to the constructor
        :return:
        """
        self.visited = {}
        self._overwrite(self.definition)

    def _overwrite(self, state: Union[states.Chain, states.State]) -> Union[states.Chain, states.State, None]:
        """
        dfs graph traversal algorithm. The logic is straightforward and is stated below:
            for a given state/chain,
            return the in-place modified input with necessary tasks replaced with cached wrappers

        :param state: Union[states.Chain, states.State], the state/chain to modify in-place
        :return: modified state
        """

        if state is None:
            return state

        # we should not visit the same node twice as the function is not idempotent
        # if a state is visited, the next visit will inject another wrapper inside the existing wrapper
        if state.state_id in self.visited:
            return self.visited[state.state_id]

        # if the input is a chain, simply iterate through it and modify all elements
        if isinstance(state, states.Chain):
            for idx, step in enumerate(state.steps):
                state.steps[idx] = self._overwrite(step)
            return state

        # if the input is a choice state, we must enter the branches as choice is not supposed to have next_step
        elif isinstance(state, states.Choice):
            for idx, (_, next_s) in enumerate(state.choices):
                # modify the branches
                state.choices[idx][1] = self._overwrite(next_s)
            # the default option is not included in state.choices
            state.default = self._overwrite(state.default)
            # ideally choice state should not have next_step. This line handles the edge case
            # where the user explicitly assigned a next_step for choice state
            # Without this line, the state machine may become disconnected
            state.next_step = self._overwrite(state.next_step)

        elif isinstance(state, states.Parallel):
            # logging.warning('Be careful that states inside any parallel state are ' +
            #                 'not cached because of a stepfunction bug')
            # iterate through the branches
            # transform them before moving forward to next_step
            for idx, next_s in enumerate(state.branches):
                state.branches[idx] = self._overwrite(next_s)
            # modify the subsequent state, otherwise it still points to the old state
            state.next_step = self._overwrite(state.next_step)

        elif isinstance(state, states.Map):
            # logging.warning('Be careful that states inside any map state are ' +
            #                 'not cached because of a stepfunction bug')
            # transform the iterator before moving forward to next_step
            state.iterator = self._overwrite(state.iterator)
            # modify the subsequent state, otherwise it still points to the old state
            state.next_step = self._overwrite(state.next_step)

        else:
            # if state is not marked for replacement, simply skip it
            # note that if a subsequent state is marked for replacement,
            # we need to fix the state.next_step pointer
            if state.state_id not in self.to_replace:
                state.next_step = self._overwrite(state.next_step)

            else:
                # note that replacement is the caching wrapper, hence it is a chain
                replacement = self.to_replace[state.state_id]
                # make sure caching_push_xxx points to the subsequent state of the input state
                replacement.steps[-1].next_step = self._overwrite(state.next_step)
                # state is inside replacement, it should not point to the original next_step
                state.next_step = None
                # mark the node as visited
                self.visited[state.state_id] = replacement
                return replacement

        # if a node is revisited, simply return the modified version
        self.visited[state.state_id] = state
        return state


class LambdaGenerator:
    CACHING_LAMBDA_DIR = 'cache_lambda'
    DEPENDENCY = 'dependency'
    DISDAT_LAYER = 'disdat_caching_layer'
    CACHING_LAMBDA_SR_PY = 'cache_lambda.py'
    LAMBDA_STUB_PY = 'lambda_for_user.py'

    @classmethod
    def generate(cls, root: Union[str, pathlib.Path] = 'generated_lambda', force_rerun: bool = False):
        """
        generate the lambda src code and a layer that users can use to create the lambda function.
        The lambda function will need access to the disdat package, which is provided as a lambda layer
        (pip install not supported by aws lambda)

        Note that installing disdat locally and zip it won't work, as some binaries are not portable. Currently
        we wrote a simple docker pip utility that installs the package in docker container (amazonlinux:2) and
        export it as a zip

        output format
            root
                - /caching_lambda
                - /dependency
                - disdat_caching_layer.zip
        :param root: str, where do you want to save the code
        :param force_rerun: bool, force rerun docker pip install even if the dependency folder is present
        :return: None
        """
        root = os.path.abspath(root)
        os.makedirs(os.path.join(root, cls.CACHING_LAMBDA_DIR), exist_ok=True)
        # copy the lambda src code. TODO: remove this after PR as caching_lambda can be accessible from disdat
        src = os.path.join(os.path.dirname(__file__), cls.CACHING_LAMBDA_SR_PY)
        dst = os.path.join(root, cls.CACHING_LAMBDA_DIR, cls.CACHING_LAMBDA_SR_PY)
        shutil.copyfile(src=src, dst=dst)
        # copy the lambda stub code that users need to create a lambda function on AWS
        src = os.path.join(os.path.dirname(__file__), cls.LAMBDA_STUB_PY)
        dst = os.path.join(root, cls.CACHING_LAMBDA_DIR, cls.LAMBDA_STUB_PY)
        shutil.copyfile(src=src, dst=dst)
        print('lambda code generated! find it in {}'.format(dst))
        # run docker pip install
        cls.install_dep(root, force_rerun)
        print("\nINSTRUCTIONS:\nPlease create a lambda function with from the generated .py file" +
              "and a lambda layer from the generated zip!\n" +
              "IMPORTANT TIPS: \nGive the lambda function proper S3 bucket permissions and timeout of at least 20s")

    @classmethod
    def install_dep(cls, root_dir: Union[str, pathlib.Path], force_rerun: bool = False):
        """
        docker pip install. Run docker build, docker run with volume mount and zip the output dependencies into a zip
        Because AWS set a size cap of 50MB for lambda layers, we did some size control by removing redundant packages
        like boto3(available by default for any lambda) as well as some dist-info folders

        :param root_dir: str, folder that is mounted with docker
        :param force_rerun: bool, re execute if true, otherwise skip if the folder already exists
        :return: None
        """
        from_folder = os.path.join(root_dir, cls.DEPENDENCY)
        to_file = os.path.join(root_dir, cls.DISDAT_LAYER)
        if not os.path.isdir(from_folder) or force_rerun:
            os.makedirs(from_folder, exist_ok=True)
            # build image from dockerfile, see the dockerfile for more details
            os.system('docker build -t docker_pip:latest {}'.format(os.path.dirname(__file__)))
            # exec container with mounted volume
            os.system('docker run -v {}:/lib/dependency docker_pip:latest'.format(from_folder))
            # zip all dependencies into a zip file
            shutil.make_archive(to_file, format='zip', root_dir=os.path.join(root_dir, cls.DEPENDENCY))
        print('layer zip generated! find it in {}'.format(to_file + '.zip'))


class StateVisitor(states.GraphVisitor):
    def __init__(self, caching: Caching):
        """
        specialized children class of GraphVisitor that determines states to replace on the fly
        :param caching: Caching, we call caching.cache_step() on selected states
        """
        self.to_replace = {}
        self.caching = caching
        super().__init__()

    def visit(self, state):
        """
        for a given state, check if we need to cache it. Note that it only make sense to cache states.Task
        :param state:
        :return:
        """
        self.states[state.state_id] = state
        if isinstance(state, states.Task):
            wrapper = self.caching.cache_step(state)
            self.to_replace[state.state_id] = wrapper
        elif isinstance(state, states.Parallel):
            for b in state.branches:
                b.accept(self)
        elif isinstance(state, states.Map):
            state.iterator.accept(self)


class ExtensiveGraphVisitor(states.GraphVisitor):
    """
    This graph visitor explores the internal structure of Map and Parallel states
    The original GraphVisitor does not support this feature
    """
    def visit(self, state):
        self.states[state.state_id] = state.to_dict()
        if isinstance(state, states.Parallel):
            for b in state.branches:
                b.accept(self)
        elif isinstance(state, states.Map):
            state.iterator.accept(self)
        else:
            pass


"""
 * Licensed to the Apache Software Foundation (ASF) under one or more
 * contributor license agreements.  See the NOTICE file distributed with
 * this work for additional information regarding copyright ownership.
 * The ASF licenses this file to You under the Apache License, Version 2.0
 * (the "License"); you may not use this file except in compliance with
 * the License.  You may obtain a copy of the License at
 *
 *    http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
"""


from disdat.common import DisdatConfig
import os
import logging
# TODO FIX THIS IMPORT ONCE THIS PLUGIN IS MERGED INTO disdat
from cache_lambda import PathParam, Cache


try:
    HOME = '/tmp/home'
    os.makedirs(HOME, exist_ok=True)
    os.environ["HOME"] = HOME
    DisdatConfig.init()
except:
    logging.warning("disdat already initialized; home overriding failed")

LOG_LEVEL = logging.INFO + 1


"""
Please copy this file to the lambda function created by you. It must have runtime environment Python>=3.8 
It also need access to Disdat, which is already packaged as a zip for you. Please create a lambda layer 
with the zip file and attach it to this lambda function 
"""


def lambda_handler(event, context):
    logging.log(level=LOG_LEVEL, msg=event)
    cache = Cache(event[PathParam.DSDT_ONLY_ARGS])
    if len(event) == 2:
        # if the input event is a dict of length 2, it's meant for cache_push
        # parent = cache.get_lineage()
        parent = None
        return cache.cache_push(event, parent)
    elif len(event) == 3:
        # if the input event is a dict of length 3, it's meant for cache_pull
        return cache.cache_pull(event)
    else:
        raise ValueError("Input event has invalid length of {}; 2/3 is expected".format(len(event)))


from setuptools import setup


def local_version(version):
    """
    Patch in a version that can be uploaded to test PyPI
    """
    return ""

setup(
    use_scm_version={
        'write_to': 'disdat_step_function/version.py',
        'write_to_template': '__version__ = "{version}"',
        'local_scheme': local_version
    },

    setup_requires=['setuptools_scm'],

    name='disdat_step_function',
    version='0.0.1',
    packages=['disdat_step_function'],
    package_dir={"": "."},

    install_requires=[
        'disdat >= 0.9.16',
        'stepfunctions == 2.2.0'
    ],
    # Choose your license
    license='Apache License, version 2.0',
    author='Ken Yocum, Zixuan Zhang',
    author_email='kyocum@gmail.com, zz2777@columbia.edu',
    url='https://github.com/kyocum/disdat-step-functions',

    extras_require={
        'dev': [
            'pytest',
            'ipython',
            'twine',
            'build'
        ],
    },

    classifiers=[
        # How mature is this project? Common values are
        #   3 - Alpha
        #   4 - Beta
        #   5 - Production/Stable
        'Development Status :: 4 - Beta',

        # Indicate who your project is intended for
        'Intended Audience :: Science/Research',
        'Topic :: Scientific/Engineering',

        # Pick your license as you wish (should match "license" above)
        'License :: OSI Approved :: Apache Software License',

        # Specify the Python versions you support here. In particular, ensure
        # that you indicate whether you support Python 2, Python 3 or both.
        'Programming Language :: Python :: 3.8',
        'Operating System :: OS Independent',
        'Natural Language :: English',
    ],
)